
/*
 * Copyright 2006, 2008 Freescale Semiconductor
 * York Sun (yorksun@freescale.com)
 * Haiying Wang (haiying.wang@freescale.com)
 * Timur Tabi (timur@freescale.com)
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <i2c.h>
#include <linux/ctype.h>

#include "../common/eeprom.h"

/**
 * static eeprom: EEPROM layout for PUID
 *
 * See application note AN3638 for details.
 */
static struct __attribute__ ((__packed__)) eeprom {
	u8 id[4];         /* 0x00 - 0x03 EEPROM Tag 'PUID' */
	u8 sn[10];        /* 0x04 - 0x0d Serial Number*/
	u8 mac[2][6];     /* 0x0e - 0x19 MAC addresses */
	u8 ip[4][4];      /* 0x1a - 0x29 IP addresses */
	u8 mask[4][4];    /* 0x2a - 0x39 Subnet Mask */
	u16 mtu[4];       /* 0x3a - 0x41 Subnet Mask */
	u8 gate[4];       /* 0x42 - 0x45 Default Gate Way */
	u8 dns[4][4];     /* 0x46 - 0x55 Domain Name Server */
	u8 host[32];      /* 0x56 - 0x75 Host Name */
	u8 domain[60];    /* 0x76 - 0xb1 Domainname */
	u16 uid;          /* 0xb2 - 0xb3 User ID */
	u16 gid;          /* 0xb4 - 0xb5 Group ID */
	u8 date[6];       /* 0xb6 - 0xbb Build Date */
	u8 errata[5];     /* 0xbc - 0xc0 Errata Level */
	u8 mac_count;     /* 0xc1        Number of MAC addresses */
	u8 res_0[58];     /* 0xc2 - 0xfb Reserved */	
	u32 crc;          /* 0xfc        CRC32 checksum */
} e;

/* Set to 1 if we've read EEPROM into memory */
static int has_been_read = 0;

/* Is this a valid CCID EEPROM? */
#define is_valid (*((u32 *)e.id) == (('P' << 24) | ('U' << 16) | ('I' << 8) | 'D'))

#define EEPROM_HOSTNAME_SIZE 32
#define EEPROM_DOMAIN_SIZE 60

/**
 * show_eeprom - display the contents of the EEPROM
 */
static void show_eeprom(void)
{
	int i;
	unsigned int crc;
#ifdef BOARD_CPU453
	unsigned char buf[256];
#endif

	/* Serial number */
	if(e.sn[0] != 0xff) printf("SN: %s\n", e.sn);

	/* Show MAC addresses  */
	for (i = 0; i < 2; i++) {
		u8 *p = e.mac[i];

		printf("eth%u: %02x:%02x:%02x:%02x:%02x:%02x\n", i,
			p[0], p[1], p[2], p[3],	p[4], p[5]);
	}

#ifdef BOARD_CPU453

	/* Show IP addresses  */
	for (i = 0; i < 4; i++) {
		u8 *p = e.ip[i];
		u8 *mp = e.mask[i];

		if(p[0] == 0xff) continue;
		printf("ip%u: %01d.%01d.%01d.%01d/%01d.%01d.%01d.%01d/%d\n",
			   i, p[0], p[1], p[2], p[3], mp[0], mp[1], mp[2], mp[3], be16_to_cpu(e.mtu[i]));
	}

	/* Show Default Gateway  */
	if(e.gate[0] != 0xff)
	printf("gateway: %01d.%01d.%01d.%01d\n", e.gate[0], e.gate[1], e.gate[2], e.gate[3]);

	/* Show Domain Name Server  */
	for (i = 0; i < 4; i++) {
		u8 *p = e.dns[i];

		if(p[0] == 0xff) continue;
		printf("dns%u: %01d.%01d.%01d.%01d\n", i, p[0], p[1], p[2], p[3]);
	}

	/* Show hostname  */
	if(e.host[0] != 0xff){
		for(i = 0; i < EEPROM_HOSTNAME_SIZE; i++) buf[i] = e.host[i];
		buf[EEPROM_HOSTNAME_SIZE] = 0;
		printf("hostname: %s\n", buf);
	}

	/* Show Domainname  */
	if(e.domain[0] != 0xff){
		for(i = 0; i < EEPROM_DOMAIN_SIZE; i++) buf[i] = e.domain[i];
		buf[EEPROM_DOMAIN_SIZE] = 0;
		printf("domainname: %s\n", buf);
	}

	/* Show User ID  */
	if(e.uid != 0xffff){
		printf("UID: %d\n", e.uid);
	}

	/* Show Group ID  */
	if(e.gid != 0xffff){
		printf("GID: %d\n", e.gid);
	}

#endif

	/* Show date, BCD date values, as YYMMDDhhmmss */
	if(e.date[0] != 0xff){
		printf("date: 20%02x/%02x/%02x %02x:%02x:%02x\n",
			   e.date[0], e.date[1], e.date[2], e.date[3], e.date[4], e.date[5]);
	}

	crc = crc32(0, (void *)&e, sizeof(e) - 4);

	if (crc == be32_to_cpu(e.crc))
		printf("CRC: %08x\n", be32_to_cpu(e.crc));
	else
		printf("CRC: %08x (should be %08x)\n",
			be32_to_cpu(e.crc), crc);

#ifdef DEBUG
	printf("EEPROM dump: (0x%x bytes)\n", sizeof(e));
	for (i = 0; i < sizeof(e); i++) {
		if ((i % 16) == 0)
			printf("%02X: ", i);
		printf("%02X ", ((u8 *)&e)[i]);
		if (((i % 16) == 15) || (i == sizeof(e) - 1))
			printf("\n");
	}
#endif
}

/**
 * read_eeprom - read the EEPROM into memory
 */
static int read_eeprom(void)
{
	int ret;
#ifdef CONFIG_SYS_EEPROM_BUS_NUM
	unsigned int bus;
#endif

	if (has_been_read)
		return 0;

#ifdef CONFIG_SYS_EEPROM_BUS_NUM
	bus = i2c_get_bus_num();
	i2c_set_bus_num(CONFIG_SYS_EEPROM_BUS_NUM);
#endif

	ret = i2c_read(CONFIG_SYS_I2C_EEPROM_ADDR, 0, CONFIG_SYS_I2C_EEPROM_ADDR_LEN,
		(void *)&e, sizeof(e));

#ifdef CONFIG_SYS_EEPROM_BUS_NUM
	i2c_set_bus_num(bus);
#endif

#ifdef DEBUG
	show_eeprom();
#endif

	has_been_read = (ret == 0) ? 1 : 0;

	return ret;
}

/**
 * prog_eeprom - write the EEPROM from memory
 */
static int prog_eeprom(void)
{
	int ret, i, length;
	unsigned int crc;
	void *p;
#ifdef CONFIG_SYS_EEPROM_BUS_NUM
	unsigned int bus;
#endif

	/* Set the reserved values to 0xFF   */
#ifdef CONFIG_SYS_I2C_EEPROM_NXID
	e.res_0 = 0xFF;
	memset(e.res_1, 0xFF, sizeof(e.res_1));
#else
	memset(e.res_0, 0xFF, sizeof(e.res_0));
#endif

	length = sizeof(e);
	crc = crc32(0, (void *)&e, length - 4);
	e.crc = cpu_to_be32(crc);

#ifdef CONFIG_SYS_EEPROM_BUS_NUM
	bus = i2c_get_bus_num();
	i2c_set_bus_num(CONFIG_SYS_EEPROM_BUS_NUM);
#endif

	for (i = 0, p = &e; i < length; i += 8, p += 8) {
		ret = i2c_write(CONFIG_SYS_I2C_EEPROM_ADDR, i, CONFIG_SYS_I2C_EEPROM_ADDR_LEN,
			p, min((length - i), 8));
		if (ret)
			break;
		udelay(5000);	/* 5ms write cycle timing */
	}

#ifdef CONFIG_SYS_EEPROM_BUS_NUM
	i2c_set_bus_num(bus);
#endif

	if (ret) {
		printf("Programming failed.\n");
		return -1;
	}

	printf("Programming passed.\n");
	return 0;
}

/**
 * h2i - converts hex character into a number
 *
 * This function takes a hexadecimal character (e.g. '7' or 'C') and returns
 * the integer equivalent.
 */
static inline u8 h2i(char p)
{
	if ((p >= '0') && (p <= '9'))
		return p - '0';

	if ((p >= 'A') && (p <= 'F'))
		return (p - 'A') + 10;

	if ((p >= 'a') && (p <= 'f'))
		return (p - 'a') + 10;

	return 0;
}

/**
 * set_date - stores the build date into the EEPROM
 *
 * This function takes a pointer to a string in the format "YYMMDDhhmmss"
 * (2-digit year, 2-digit month, etc), converts it to a 6-byte BCD string,
 * and stores it in the build date field of the EEPROM local copy.
 */
static void set_date(const char *string)
{
	unsigned int i;

	if(!string){
		for (i = 0; i < 6; i++) e.date[i] = 0xff;
		return;
	}

	if (strlen(string) != 12) {
		printf("Usage: mac date YYMMDDhhmmss\n");
		return;
	}

	for (i = 0; i < 6; i++)
		e.date[i] = h2i(string[2 * i]) << 4 | h2i(string[2 * i + 1]);
}

/**
 * set_mac_address - stores a MAC address into the EEPROM
 *
 * This function takes a pointer to MAC address string
 * (i.e."XX:XX:XX:XX:XX:XX", where "XX" is a two-digit hex number) and
 * stores it in one of the MAC address fields of the EEPROM local copy.
 */
static void set_mac_address(unsigned int index, const char *string)
{
	char *p = (char *) string;
	unsigned int i;

	if (!string) {
		printf("Usage: mac <n> XX:XX:XX:XX:XX:XX\n");
		return;
	}

	for (i = 0; *p && (i < 6); i++) {
		e.mac[index][i] = simple_strtoul(p, &p, 16);
		if (*p == ':')
			p++;
	}
}

#ifdef CONFIG_SYS_I2C_EEPROM_PUID
static void set_def_gateway(unsigned int index, const char *string)
{
	char *p = (char *) string;
	unsigned int i;

	if (!string){
		for (i = 0; i < 4; i++) e.gate[i] = 0xff;
		return;
	}

	for (i = 0; *p && (i < 4); i++) {
		e.gate[i] = simple_strtoul(p, &p, 10);
		if (*p == '.') p++;
	}
}
static void set_mtu(unsigned int index, const char *string)
{
	char *p = (char *) string;

	if (!string){
		e.mtu[index] = 0xffff;
		return;
	}

	e.mtu[index] = min(simple_strtoul(p, &p, 10), 9000);
}

static void set_subnet_mask(unsigned int index, const char *string)
{
	char *p = (char *) string;
	unsigned int i;

	if (index >= 4) return;
	if (!string){
		for (i = 0; i < 4; i++) e.mask[index][i] = 0xff;
		return;
	}

	for (i = 0; *p && (i < 4); i++) {
		e.mask[index][i] = simple_strtoul(p, &p, 10);
		if (*p == '.') p++;
	}
}
static void set_ip_address(unsigned int index, const char *string)
{
	char *p = (char *) string;
	unsigned int i;

	if (index >= 4) return;
	if (!string) {
		for (i = 0; i < 4; i++) e.ip[index][i] = 0xff;
		return;
	}

	for (i = 0; *p && (i < 4); i++) {
		e.ip[index][i] = simple_strtoul(p, &p, 10);
		if (*p == '.') p++;
	}
	if((p = strchr(p, '/')) != NULL) set_subnet_mask(index, ++p);
	if((p = strchr(p, '/')) != NULL) set_mtu(index, ++p);
}

static void set_dns_address(unsigned int index, const char *string)
{
	char *p = (char *) string;
	unsigned int i;

	if (index >= 4) return;
	if (!string){
		for (i = 0; i < 4; i++) e.dns[index][i] = 0xff;
		return;
	}

	for (i = 0; *p && (i < 4); i++) {
		e.dns[index][i] = simple_strtoul(p, &p, 10);
		if (*p == '.') p++;
	}
}

static void set_uid(unsigned int index, const char *string)
{
	char *p = (char *) string;

	if (!string){
		e.uid = 0xffff;
		return;
	}
	e.uid = min(simple_strtoul(p, &p, 10), 0xffff);
	if(e.uid < 500) e.uid = 500;
}

static void set_gid(unsigned int index, const char *string)
{
	char *p = (char *) string;

	if (!string){
		e.gid = 0xffff;
		return;
	}
	e.gid = min(simple_strtoul(p, &p, 10), 0xffff);
	if(e.gid < 500) e.gid = 500;
}

#endif

#define MAC_CMD_READ (('r'<<24)+('e'<<16)+('a'<<8)+'d')
#define MAC_CMD_SAVE (('s'<<24)+('a'<<16)+('v'<<8)+'e')
#define MAC_CMD_SN (('s'<<8)+'n')
#define MAC_CMD_ERR (('e'<<16)+('r'<<8)+'r')
#define MAC_CMD_DATE (('d'<<24)+('a'<<16)+('t'<<8)+'e')
#define MAC_CMD_PORT (('p'<<24)+('o'<<16)+('r'<<8)+'t')
#define MAC_CMD_MAC (('m'<<16)+('a'<<8)+'c')
#define MAC_CMD_IP (('i'<<8)+'p')
#define MAC_CMD_SUB (('s'<<16)+('u'<<8)+'b')
#define MAC_CMD_MTU (('m'<<16)+('t'<<8)+'u')
#define MAC_CMD_GATE (('g'<<24)+('a'<<16)+('t'<<8)+'e')
#define MAC_CMD_DNS (('d'<<16)+('n'<<8)+'s')
#define MAC_CMD_HOST (('h'<<24)+('o'<<16)+('s'<<8)+'t')
#define MAC_CMD_DOM (('d'<<16)+('o'<<8)+'m')
#define MAC_CMD_UID (('u'<<16)+('i'<<8)+'d')
#define MAC_CMD_GID (('g'<<16)+('i'<<8)+'d')

int do_mac(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int i, cmd, len;

	if (argc == 1) {
		show_eeprom();
		return 0;
	}

	len = strlen(argv[1]);
	if(len > 4) len = 4;

	for(i = 0, cmd = 0; i < len; i++){
		cmd <<= 8;
		cmd += argv[1][i];
	}


	if (cmd == MAC_CMD_READ) {
		read_eeprom();
		return 0;
	}

	if (!is_valid) {
		(*((u32 *)e.id) = (('P' << 24) | ('U' << 16) | ('I' << 8) | 'D'));
		e.mac_count = 2;
	}

	switch (cmd) {
	case MAC_CMD_SAVE:	/* save */
		prog_eeprom();
		break;
	case MAC_CMD_SN:	/* serial number */
		memset(e.sn, 0xff, sizeof(e.sn));
		if(argv[2]) strncpy((char *)e.sn, argv[2], sizeof(e.sn) - 1);
		break;
	case MAC_CMD_ERR:	/* errata */
		memset(e.errata, 0xff, 5);
		if(argv[2]) strncpy((char *)e.errata, argv[2], 4);
		break;
	case MAC_CMD_DATE:	/* date BCD format YYMMDDhhmmss */
		set_date(argv[2]);
		break;
	case MAC_CMD_PORT:	/* MAC table size */
		e.mac_count = simple_strtoul(argv[2], NULL, 16);
		break;
	case '0' ... '1':	/* "mac 0" through "mac 1" */
		set_mac_address(argv[1][0] - '0', argv[2]);
		break;
	case MAC_CMD_MAC:	/* "mac 0" through "mac 1" */
#ifdef BOARD_CPU453
		set_mac_address(argv[2][0] - '0', argv[3]);
#endif
		break;
	case MAC_CMD_IP:	/* ip addr */
#ifdef BOARD_CPU453
		set_ip_address(argv[2][0] - '0', argv[3]);
#endif
		break;
	case MAC_CMD_SUB:	/* subnet mask */
#ifdef BOARD_CPU453
		set_subnet_mask(argv[2][0] - '0', argv[3]);
#endif
		break;
	case MAC_CMD_MTU:	/* mtu */
#ifdef BOARD_CPU453
		set_mtu(argv[2][0] - '0', argv[3]);
#endif
		break;
	case MAC_CMD_GATE:	/* gateway */
//		set_def_gateway('0', argv[2]);
		break;
	case MAC_CMD_DNS:	/* domain name server */
//		set_dns_address(argv[2][0] - '0', argv[3]);
		break;
	case MAC_CMD_HOST:	/* hostname */
//		memset(e.host, 0xff, sizeof(e.host));
//		if(argv[2]) strncpy((char *)e.host, argv[2], sizeof(e.host));
		break;
	case MAC_CMD_DOM:	/* domainname */
//		memset(e.domain, 0xff, sizeof(e.domain));
//		if(argv[2]) strncpy((char *)e.domain, argv[2], sizeof(e.domain));
		break;
	case MAC_CMD_UID:	/* uid */
//		set_uid('0', argv[2]);
		break;
	case MAC_CMD_GID:	/* gid */
//		set_gid('0', argv[2]);
		break;
	default:
		cmd_usage(cmdtp);
		break;
	}

	return 0;
}

/**
 * mac_read_from_eeprom - read the MAC addresses from EEPROM
 *
 * This function reads the MAC addresses from EEPROM and sets the
 * appropriate environment variables for each one read.
 *
 * The environment variables are only set if they haven't been set already.
 * This ensures that any user-saved variables are never overwritten.
 *
 * This function must be called after relocation.
 */
int mac_read_from_eeprom(void)
{
	unsigned int i;

	if (read_eeprom()) {
		printf("Read failed.\n");
		return -1;
	}

	if (!is_valid) {
		printf("Invalid ID (%02x %02x %02x %02x)\n", e.id[0], e.id[1], e.id[2], e.id[3]);
		return -1;
	}

	if (be32_to_cpu(e.crc) != 0xFFFFFFFF) {
		u32 crc = crc32(0, (void *)&e, sizeof(e) - 4);

		if (crc != be32_to_cpu(e.crc)) {
			printf("CRC mismatch (%08x != %08x).\n", crc,
				be32_to_cpu(e.crc));
			return -1;
		}
	}

	for (i = 0; i < min(4, e.mac_count); i++) {
		if (memcmp(&e.mac[i], "\0\0\0\0\0\0", 6) &&
		    memcmp(&e.mac[i], "\xFF\xFF\xFF\xFF\xFF\xFF", 6)) {
			char ethaddr[18];
			char enetvar[9];

			sprintf(ethaddr, "%02X:%02X:%02X:%02X:%02X:%02X",
				e.mac[i][0],
				e.mac[i][1],
				e.mac[i][2],
				e.mac[i][3],
				e.mac[i][4],
				e.mac[i][5]);
			sprintf(enetvar, i ? "eth%daddr" : "ethaddr", i);
			/* Only initialize environment variables that are blank
			 * (i.e. have not yet been set)
			 */
			if (!getenv(enetvar))
				setenv(enetvar, ethaddr);
			debug("I2C: %s: %s\n",enetvar, ethaddr);
		}
	}

#ifdef CONFIG_SYS_I2C_EEPROM_PUID
	for (i = 0; i < 4; i++) {
		if (memcmp(&e.ip[i], "\0\0\0\0", 4) &&
		    memcmp(&e.ip[i], "\xFF\xFF\xFF\xFF", 4)) {
			char ipaddr[16];
			char ipvar[9];

			sprintf(ipaddr, "%01d.%01d.%01d.%01d",
				e.ip[i][0],
				e.ip[i][1],
				e.ip[i][2],
				e.ip[i][3]);
			sprintf(ipvar, i ? "ip%daddr" : "ipaddr", i);
			/* Only initialize environment variables that are blank
			 * (i.e. have not yet been set)
			 */
			if (!getenv(ipvar))
				setenv(ipvar, ipaddr);
			debug("I2C: %s: %s\n",ipvar, ipaddr);
		}
	}
#endif

	return 0;
}

#ifdef CONFIG_SYS_I2C_EEPROM_CCID

/**
 * get_cpu_board_revision - get the CPU board revision on 85xx boards
 *
 * Read the EEPROM to determine the board revision.
 *
 * This function is called before relocation, so we need to read a private
 * copy of the EEPROM into a local variable on the stack.
 *
 * Also, we assume that CONFIG_SYS_EEPROM_BUS_NUM == CONFIG_SYS_SPD_BUS_NUM.  The global
 * variable i2c_bus_num must be compile-time initialized to CONFIG_SYS_SPD_BUS_NUM,
 * so that the SPD code will work.  This means that all pre-relocation I2C
 * operations can only occur on the CONFIG_SYS_SPD_BUS_NUM bus.  So if
 * CONFIG_SYS_EEPROM_BUS_NUM != CONFIG_SYS_SPD_BUS_NUM, then we can't read the EEPROM when
 * this function is called.  Oh well.
 */
unsigned int get_cpu_board_revision(void)
{
	struct board_eeprom {
		u32 id;           /* 0x00 - 0x03 EEPROM Tag 'CCID' */
		u8 major;         /* 0x04        Board revision, major */
		u8 minor;         /* 0x05        Board revision, minor */
	} be;

	i2c_read(CONFIG_SYS_I2C_EEPROM_ADDR, 0, CONFIG_SYS_I2C_EEPROM_ADDR_LEN,
		(void *)&be, sizeof(be));

	if (be.id != (('C' << 24) | ('C' << 16) | ('I' << 8) | 'D'))
		return MPC85XX_CPU_BOARD_REV(0, 0);

	if ((be.major == 0xff) && (be.minor == 0xff))
		return MPC85XX_CPU_BOARD_REV(0, 0);

	return MPC85XX_CPU_BOARD_REV(be.major, be.minor);
}
#endif
